module Network.RPCA.Connection
  ( Connection
  , new
  , forkThreads
  , readPacket
  , writePacket
  , buildHeader
  , connoutq
  ) where

import Control.Concurrent
import Control.Concurrent.STM
import Control.Exception
import Control.Monad

import Data.Binary.Strict.Get
import Data.Binary.Put
import Foreign.C.Types

import Data.Maybe (fromJust)
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import qualified Data.Map as Map
import qualified Data.Sequence as Seq

import Network.Socket hiding (send, sendTo, recv, recvFrom)
import Network.Socket.ByteString

import Network.RPCA.Structs

data Connection = Connection { connsocket :: Socket
                             , connoutq :: TVar (Seq.Seq BS.ByteString)
                             , connreaderthread :: TVar (Maybe ThreadId)
                             , connwriterthread :: TVar (Maybe ThreadId)
                             , conndeath :: IO ()
                             , conndead :: TVar Bool }

-- Threading: each connection has two threads: one pumping data out and one
-- reading it in. The failure (by throwing an exception) of either is enough
-- to close the socket and kill the connection. This could happen to both at
-- the same time, in which case they race to set conndead to True.
-- However, we also need to record their ThreadIds in the Connection record
-- so that they can kill each other. It's possible that they could try to
-- kill the connection right away - before the creating thread has recorded
-- the correct ThreadIds in the Connection. Thus, at startup, they both wait
-- for the controlling thread to set conndead to False before doing anything
-- else

new :: Socket  -- ^ the socket to make a connection from
    -> IO ()  -- ^ the action run when the connection fails
    -> STM Connection
new socket deathaction = do
  dead <- newTVar True
  outq <- newTVar Seq.empty
  p1 <- newTVar Nothing
  p2 <- newTVar Nothing

  let conn = Connection socket outq p1 p2 deathaction dead
  return conn

forkThreads :: Connection  -- ^ the connection to fork the threads for
            -> IO ()  -- ^ the action which reads from the socket
            -> IO ()
forkThreads conn readeraction = do
  reader <- forkIO $ waitForReadySignal conn $
                     connectionThreadWrapper conn connwriterthread $
                     readeraction
  writer <- forkIO $ waitForReadySignal conn $
                     connectionThreadWrapper conn connreaderthread $
                     seqToSocket (connoutq conn) (connsocket conn)
  -- update the thread ids in the Connection and set the ready flag
  atomically (writeTVar (connreaderthread conn) (Just reader) >>
              writeTVar (connwriterthread conn) (Just writer) >>
              writeTVar (conndead conn) False)
  return ()

-- | Wait for conndead to be set to False on the given connection, then run
--   the given action
waitForReadySignal :: Connection -> IO a -> IO a
waitForReadySignal conn action = do
  atomically (do dead <- readTVar (conndead conn)
                 if dead == True then retry else return ())
  action

-- | Wrap a connection thread so that, when the thread dies, it races to set
--   the dead flag. If it does so, it closes the socket and kills the other
--   thread
connectionThreadWrapper :: Connection -> (Connection -> TVar (Maybe ThreadId)) -> IO a -> IO a
connectionThreadWrapper conn otherthread action = do
  finally action
          (do isDead <- atomically (do dead <- readTVar (conndead conn)
                                       when (not dead) $ writeTVar (conndead conn) True
                                       return dead)
              when (not isDead) (do t <- atomically (readTVar $ otherthread conn)
                                    killThread $ fromJust t
                                    sClose (connsocket conn)
                                    conndeath conn))

-- | Atomically take elements from the end of the given sequence and write them
--   to the given socket. Throw an exception when the write fails
seqToSocket :: TVar (Seq.Seq BS.ByteString)  -- ^ data is removed from the end
            -> Socket  -- ^ the socket to write to
            -> IO ()
seqToSocket q sock = do
  -- Atomically remove an element from the end of the sequence
  bs <- atomically (do q' <- readTVar q
                       (bs, rest) <-
                         case Seq.viewr q' of
                              Seq.EmptyR -> retry
                              rest Seq.:> head -> return (head, rest)
                       writeTVar q rest
                       return bs)
  -- Write the data to the socket
  bytes <- send sock bs
  when (bytes /= BS.length bs) $ error "Short write"
  seqToSocket q sock

-- | Read a given number of bytes from a socket. Block until that many bytes
--   are availible. Throw an exception otherwise
reada :: Socket -> Int -> IO BS.ByteString
reada _ 0 = return BS.empty
reada sock bytes = do
  bs <- recv sock bytes
  if BS.length bs < bytes
     then if BS.null bs
             then error "Disconnected"
             else reada sock (bytes - BS.length bs) >>= return . BS.append bs
     else return bs

kMaxSerialisedRPCHeaderSize = 512
kMaxSerialisedPayloadSize = 1024 * 1024

-- | Read an rpc header and payload from the wire.
readPacket :: Socket -> IO (BS.ByteString, BS.ByteString)
readPacket sock = do
  -- Read the 12 byte header
  header <- reada sock 12
  -- Parse the header from the wire
  let deserialiseHeader = do
        alen <- getWord32be
        plen <- getWord32be
        chksum <- getWord32be
        return (alen, plen, chksum)
  case fst (runGet deserialiseHeader header) of
       Left _ -> error "Protocol error in header"
       Right (alen, plen, chksum) -> do
         when (alen > kMaxSerialisedRPCHeaderSize) $ error "RPC header too large"
         when (plen > kMaxSerialisedPayloadSize) $ error "Payload too large"
         -- Read the rpcheader and payload
         a <- reada sock $ fromIntegral alen
         payload <- reada sock $ fromIntegral plen
         return (a, payload)

writePacket :: Connection -> (BS.ByteString, BS.ByteString) -> STM ()
writePacket conn (a, payload) = do
  let header = buildHeader (a, payload)
  q <- readTVar $ connoutq conn
  writeTVar (connoutq conn) $ payload Seq.<| a Seq.<| header Seq.<| q

-- | Build a 12-byte header for the given ib/ob header and payload
buildHeader :: (BS.ByteString, BS.ByteString) -> BS.ByteString
buildHeader (a, b) =
  BS.concat $ BSL.toChunks $ runPut (do
    putWord32be $ fromIntegral $ BS.length a
    putWord32be $ fromIntegral $ BS.length b
    putWord32be 0)
